#include "world.h"
#include "core/character.h"
#include "core/renderer.h"
#include "core/roguedefs.h"
#include <SDL2/SDL_log.h>
#include <cassert>
#include <memory>

namespace rogue {
void Room::draw(Tile world_pivot) {
	for(Tile local_tile{0, 0}; local_tile.y < this->chunk_size;) {
		if(this->tile_is_wall({local_tile.x, local_tile.y})) {
			Render::draw({
				world_pivot.x + local_tile.x,
				world_pivot.y + local_tile.y
			}, 1);
		}
		if((++local_tile.x %= this->chunk_size) == 0) {
			++local_tile.y;
		}
	}
}

bool Room::tile_is_wall(Tile local_tile) const {
	bool const is_outside_room{
		local_tile.x <= this->rect.x
		|| local_tile.y <= this->rect.y
		|| local_tile.x >= this->rect.x + this->rect.w-1
		|| local_tile.y >= this->rect.y + this->rect.h-1};
	int chunk_size2 = chunk_size/2;
	bool const is_outside_hallways{
		((this->hallway_paths & Directions::NORTH) == 0 || local_tile.x != chunk_size / 2 || local_tile.y > chunk_size2)
		&& ((this->hallway_paths & Directions::SOUTH) == 0 || local_tile.x != chunk_size / 2 || local_tile.y < chunk_size2)
		&& ((this->hallway_paths & Directions::WEST) == 0 || local_tile.y != chunk_size / 2 || local_tile.x < chunk_size2)
		&& ((this->hallway_paths & Directions::EAST) == 0 || local_tile.y != chunk_size / 2 || local_tile.x > chunk_size2)
	};
	return is_outside_room && is_outside_hallways;
}

World::World()
: direction_pressed{{
		SDL_SCANCODE_H, SDL_SCANCODE_J, SDL_SCANCODE_K, SDL_SCANCODE_L,
		SDL_SCANCODE_LEFT, SDL_SCANCODE_RIGHT, SDL_SCANCODE_UP, SDL_SCANCODE_DOWN
	},
	std::bind(&World::on_input, this, std::placeholders::_1)
} {}

void World::act() {
	for(Character &character : this->characters) {
		character.act();
	}
}

void World::render() {
	Render::clear(this->player->location);
	for(size_t i{0}; i < this->rooms.size(); ++i) {
		this->rooms[i].draw({
			.x = int(i % this->shear * this->chunk_size),
			.y = int(i / this->shear * this->chunk_size)
		});
	}
	for(Character &character : this->characters) {
		character.draw();
	}
}

Room &World::get_room(Chunk chunk) {
	return rooms[chunk.x + chunk.y * this->shear];
}

TileData World::query_tile(Tile tile) {
	Room &room{this->get_room({tile.x / this->chunk_size, tile.y / this->chunk_size})};
	TileData out{
		.character = nullptr,
		.is_wall = room.tile_is_wall({tile.x % this->chunk_size, tile.y % this->chunk_size})
	};
	for(Character &character : this->characters) {
		if(character.location == tile) {
			out.character = &character;
			return out;
		}
	}
	return out;
}

void World::on_input(bool pressed) {
	if(!pressed) {
		this->act();
	}
}

WorldGenerator &WorldGenerator::with_chunk_size(unsigned side_length) {
	assert(this->chunk_side_length == 0);
	assert(side_length > 0);
	this->chunk_side_length = side_length;
	return *this;
}

WorldGenerator &WorldGenerator::with_world_size(unsigned side_length) {
	assert(this->world_side_length == 0);
	assert(side_length > 0);
	this->world_side_length = side_length;
	return *this;
}

WorldGenerator &WorldGenerator::with_room_size(unsigned min_side_length, unsigned max_side_length) {
	assert(chunk_side_length > 0 && "Chunk size needs to be initialized before room size");
	assert(min_side_length < max_side_length);
	assert(min_side_length > 0);
	this->min_room_size = min_side_length;
	this->max_room_size = max_side_length;
	return *this;
}

WorldGenerator &WorldGenerator::with_player(Character character) {
	this->player.emplace(character);
	return *this;
}

std::unique_ptr<World> WorldGenerator::generate() {
	assert(this->chunk_side_length > 0 && "Chunk size is required to generate world");
	assert(this->min_room_size > 0 && "Room size is required to generate world");
	assert(this->min_room_size < this->max_room_size && "Room size is required to generate world");
	assert(this->world_side_length > 0 && "World size is required to generate world");
	assert(this->player.has_value() && "World cannot be generated without a player"); 
	std::unique_ptr<World> world{new World()};
	world->characters.push_back(player.value());
	world->player = &world->characters[0];
	world->player->world = world.get();
	world->chunk_size = this->chunk_side_length;
	world->shear = this->world_side_length;
	world->rooms.reserve(this->world_side_length * this->world_side_length);
	world->rooms.push_back(Room {
		.hallway_paths = Directions(Directions::NORTH | Directions::EAST),
		.chunk_size = this->chunk_side_length,
		.rect = {
			(this->chunk_side_length - this->max_room_size) / 2,
			(this->chunk_side_length - this->max_room_size) / 2,
			this->max_room_size,
			this->max_room_size
		}
	});
	return world;
}
}
